;;
;;

(define m17nlib-candidate-max 10)

(define m17nlib-context-id
  (lambda (mc)
    (car (nthcdr 0 mc))))

(define m17nlib-context-set-id!
  (lambda (mc id)
    (set-car! (nthcdr 0 mc) id)))

(define m17nlib-context-showing-candidate
  (lambda (mc)
    (car (nthcdr 1 mc))))

(define m17nlib-context-set-showing-candidate!
  (lambda (mc id)
    (set-car! (nthcdr 1 mc) id)))

(define m17nlib-context-new
  (lambda (name)
    (let ((c ()))
      (set! c (copy-list '(#f #f)))
      (m17nlib-context-set-id! c
			       (m17nlib-lib-alloc-context name))
      c)))

(define m17nlib-update-preedit
  (lambda (id)
    (let* ((c (find-context id))
	   (mc (context-data c))
	   (mid (m17nlib-context-id mc)))
      (if (m17nlib-lib-preedit-changed? mid)
	  (if (m17nlib-lib-compose-mode? mid)
	      (begin
		(im-clear-preedit id)
		(im-pushback-preedit id
				     preedit-underline
				     (m17nlib-lib-get-left-of-candidate mid))
		(im-pushback-preedit id
				     (+ preedit-reverse preedit-cursor)
				     (m17nlib-lib-get-selected-candidate mid))
		(im-pushback-preedit id
				     preedit-underline
				     (m17nlib-lib-get-right-of-candidate mid))
		(im-update-preedit id))
	      (begin
		(im-clear-preedit id)
		(im-pushback-preedit id
				     preedit-underline
				     (m17nlib-lib-get-left-of-cursor mid))
		(im-pushback-preedit id
				     preedit-cursor "")
		(im-pushback-preedit id
				     preedit-underline
				     (m17nlib-lib-get-right-of-cursor mid))
		(im-update-preedit id))
	      )))))

(define m17nlib-update-candidate
  (lambda (id)
    (let* ((c (find-context id))
	   (mc (context-data c))
	   (mid (m17nlib-context-id mc))
	   (showing-candidate? (m17nlib-context-showing-candidate mc)))
      (if showing-candidate?
	  (if (m17nlib-lib-candidate-show? mid)
	      (begin
		(im-select-candidate mid
				     (m17nlib-lib-get-candidate-index mid)))
	      (begin
		(m17nlib-context-set-showing-candidate! mc #f)
		(im-deactivate-candidate-selector id mid))) ;end showing candidate
	  (if (m17nlib-lib-candidate-show? mid)
	      (begin
		(im-activate-candidate-selector
		 id
		 5 m17nlib-candidate-max)
		(im-select-candidate id (m17nlib-lib-get-candidate-index mid))
		(m17nlib-context-set-showing-candidate! mc #t)))
	  ))));nothing to do

(define m17nlib-append-modifiers
  (lambda (key key-state key-str)
    (if (shift-key-mask key-state)
	(set! key-str (string-append "S-" key-str)))
    (if (control-key-mask key-state)
	(set! key-str (string-append "C-" key-str)))
    (if (alt-key-mask key-state)
	(set! key-str (string-append "A-" key-str)))
    (if (meta-key-mask key-state)
	(set! key-str (string-append "M-" key-str)))
    key-str))

; Unfortunatelly, we don't have simple way to translate...
(define m17nlib-translate-ukey-to-mkey
  (lambda (key key-state)
    (m17nlib-append-modifiers
     key key-state
     (cond     
      ((= key 'backspace)
       "BackSpace")
      ((= key 'delete)
       "Delete")
      ((= key 'escape)
       "Escape")
      ((= key 'return)
       "Return")
      ((= key 'tab)
       "Tab")
      ((= key 'left)
       "Left")
      ((= key 'up)
       "Up")
      ((= key 'right)
       "Right")
      ((= key 'down)
       "Down")
      ((= key 'prior)
       "Page_Down")
      ((= key 'next)
       "Page_Up")
      ((= key 'home)
       "Home")
      ((= key 'end)
       "End")
      ((= key 'zenkaku-hankaku)
       "")
      ((= key 'Multi_key)
       "")
      ((= key 'Mode_switch)
       "")
      ((= key 'Henkan_Mode)
       "")
      ((= key 'Muhenkan)
       "")
      ((= key 'F1)
       "F1")
      ((= key 'F2)
       "F2")
      ((= key 'F3)
       "F3")
      ((= key 'F4)
       "F4")
      ((= key 'F5)
       "F5")
      ((= key 'F6)
       "F6")
      ((= key 'F7)
       "F7")
      ((= key 'F8)
       "F8")
      ((= key 'F9)
       "F9")
      ((= key 'F10)
       "F10")
      ((= key 'F11)
       "F11")
      ((= key 'F12)
       "F12")
      ((= key 'F13)
       "F13")
      ((= key 'F14)
       "F14")
      ((= key 'F15)
       "F15")
      ((= key 'F16)
       "F16")
      ((= key 'F17)
       "F17")
      ((= key 'F18)
       "F18")
      ((= key 'F19)
       "F19")
      ((= key 'F20)
       "F20")))))
  
(define m17nlib-init-handler
  (lambda (id arg)
    (let* ((c (find-context id)))
      (set-context-data! c (m17nlib-context-new arg)))))

(define m17nlib-release-handler
  (lambda (id)))

(define m17nlib-mode-handler
  (lambda (id mode)))

(define m17nlib-push-key
  (lambda (id key key-state)
    (let* ((c (find-context id))
	   (mc (context-data c))
	   (mid (m17nlib-context-id mc)))
      (if (symbol? key)
	  (let ((mkey (m17nlib-translate-ukey-to-mkey key key-state)))
	    (m17nlib-lib-push-symbol-key mid mkey))
	  (m17nlib-lib-push-key mid key key-state)))))

(define m17nlib-press-key-handler
  (lambda (id key key-state)
    (let* ((c (find-context id))
	   (mc (context-data c))
	   (mid (m17nlib-context-id mc)))
      (if (m17nlib-push-key id key key-state)
	  #f ;; Discard key event
	  (let ((commit-str (m17nlib-lib-get-commit-string mid)))
	    (if (string=? commit-str "")
		(im-commit-raw id)
		(begin
		  (im-commit id commit-str)
		  (m17nlib-lib-commit mid)))))
      (m17nlib-update-preedit id)
      (m17nlib-update-candidate id))))

(define m17nlib-release-key-handler
  (lambda (id key key-state)
    ()))

(define m17nlib-reset-handler
  (lambda (id)
    ()))

(define m17nlib-get-candidate-handler
  (lambda (id idx accel-enum-hint)
    (let* ((c (find-context id))
	   (mc (context-data c))
	   (mid (m17nlib-context-id mc))
	   (cand (m17nlib-lib-get-nth-candidate
		  mid idx)))
      (list cand (digit->string (+ idx 1))))))

(define m17nlib-set-candidate-index-handler
  (lambda (id idx)))

(define m17nlib-prop-handler
  (lambda (id message)))

(define m17nlib-register
  (lambda (i nr-im)
    (if (> nr-im i)
	(begin
	  (register-im
	   (m17nlib-lib-nth-input-method-name i)
	   (m17nlib-lib-nth-input-method-lang i)
	   "UTF-8"
	   (m17nlib-lib-nth-input-method-name i)
	   m17nlib-init-handler
	   m17nlib-release-handler
	   m17nlib-mode-handler
	   m17nlib-press-key-handler
	   m17nlib-release-key-handler
	   m17nlib-reset-handler
	   m17nlib-get-candidate-handler
	   m17nlib-set-candidate-index-handler
	   m17nlib-prop-handler
	   )
	  (m17nlib-register (+ i 1) nr-im))
	())))

(m17nlib-lib-init)
(m17nlib-register 0 (m17nlib-lib-nr-input-methods))
